/*
 * Copyright (C) 2009 HB1 <http://www.h7.dion.ne.jp/~rasiel/zega.html>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "custommaplist.h"

using namespace std;

CustomMapList::CustomMapList() {
}

CustomMapList::~CustomMapList() {
	_latitude.clear();
	_longitude.clear();
	if (_mapListFile.isOpen())
		_mapListFile.close();
}

void CustomMapList::changeMap(const QString &path) {
	// ե륯
	if (_mapListFile.isOpen())
		_mapListFile.close();

	// 
	_maxFileLen = 0;
	_basePath = path + "/";
	_mapListFile.setName(path + "/maplist.dat");

	if (_mapListFile.exists()) {
		if (_mapListFile.open(IO_ReadOnly)) {
			QDataStream ds(&_mapListFile);
			Q_UINT32 magic;
			Q_UINT32 latCount, lngCount;
			Q_INT32 readData;
			ds >> magic;
			if (magic == (Q_UINT32)0x15614878) {
				ds >> _maxFileLen;
				ds >> latCount;
				ds >> lngCount;

				_baseLength = (4 + latCount + lngCount) * 4;

				_latitude.clear();
				_longitude.clear();
				for (unsigned int i = 0; i < latCount; ++i) {
					ds >> readData;
					_latitude.insert(pair<long, int>(readData, i));
				}
				for (unsigned int i = 0; i < lngCount; ++i) {
					ds >> readData;
					_longitude.insert(pair<long, int>(readData, i));
				}
			} else {
				// File Type Error.
			}
			_mapListFile.close();
		}
	} else {
		_latitude.clear();
		_longitude.clear();
		createMapList();
	}

	if (_mapListFile.open(IO_ReadOnly)) {
		cout << "file size  " << _mapListFile.size() << endl;
		cout << "baseLength " << _baseLength << endl;
		cout << "latitude   " << _latitude.size() << endl;
		cout << "longitude  " << _longitude.size() << endl;
	}

	_latid = _longid = 0;
}

void CustomMapList::createMapList() {
	map<long, int>::iterator it;
	int indexNum;
	char tempString[1024];

	// ٰ٤ΰ
	cout << "check itude" << endl;
	updateItudeList("", &_latitude, &_longitude, &_maxFileLen);

	_maxFileLen++;
	for (unsigned int i = 0; i < _maxFileLen; i++)
		tempString[i] = '\0';

	// ޥåץǥå
	if (_mapListFile.open(IO_WriteOnly)) {
		cout << "write maplist file: " << _mapListFile.name() << endl;
		QDataStream ds(&_mapListFile);
		ds << (Q_UINT32)0x15614878; //_magicNumber;
		ds << (Q_UINT32)_maxFileLen;
		ds << (Q_UINT32)_latitude.size();
		ds << (Q_UINT32)_longitude.size();

		_baseLength = 16;

		// ٰ
		cout << "latitude list" << endl;
		it = _latitude.begin();
		indexNum = 0;
		while (it != _latitude.end()) {
			ds << (Q_INT32)it->first;
			it->second = indexNum++;
			_baseLength += 4;
			++it;
		}

		// ٰ
		cout << "longitude list" << endl;
		it = _longitude.begin();
		indexNum = 0;
		while (it != _longitude.end()) {
			ds << (Q_INT32)it->first;
			it->second = indexNum++;
			_baseLength += 4;
			++it;
		}

		// mapե̾񤭽Ф
		cout << "size " << _latitude.size() * _longitude.size() << endl;
		QTextStream ts(&_mapListFile);
		for (unsigned int i = 0; i < _latitude.size() * _longitude.size(); i++)
			ts.writeRawBytes(tempString, _maxFileLen);

		updateMapList("", &_mapListFile);
	}

	_mapListFile.close();
}

void CustomMapList::updateItudeList(const QString &path,
		map<long, int> *latitude, map<long, int> *longitude,
		unsigned int * maxFileLen) {
	QDir dir(_basePath + path);
	QStringList::Iterator it;
	Radian radNS, radEW;
	unsigned int maxDirFileLen = 0;

	// Ͽ޲եٰ̾٤ΰ
	QStringList files = dir.entryList("*.bmp *.gif", QDir::Files);
	it = files.begin();
	while (it != files.end()) {
		QStringList mapInfo = QStringList::split(",", *it);
		if (radNS.parse(mapInfo[0]) && radEW.parse(mapInfo[1])) {
			latitude->insert(pair<long, int>(radNS.toLong(), 0));
			longitude->insert(pair<long, int>(radEW.toLong(), 0));
			// ե륵
			if ((*it).length() > *maxFileLen)
				*maxFileLen = (*it).length();
		}
		++it;
	}

	// ֥ǥ쥯ȥ⸡
	QStringList dirs = dir.entryList(QDir::Dirs);
	it = dirs.begin();
	while (it != dirs.end()) {
		if (*it != "." && *it != "..") {
			updateItudeList(path + *it + "/", latitude, longitude,
					&maxDirFileLen);
			maxDirFileLen = (*it).length() + 1 + maxDirFileLen;
			if (maxDirFileLen > *maxFileLen)
				*maxFileLen = maxDirFileLen;
		}
		++it;
	}
}

void CustomMapList::updateMapList(const QString &path, QFile *mapListFile) {
	QDir dir(_basePath + path);
	int latid, longid;
	QStringList::Iterator it;
	Radian radNS, radEW;

	QStringList files = dir.entryList("*.bmp *.gif", QDir::Files);
	it = files.begin();
	while (it != files.end()) {
		QStringList mapInfo = QStringList::split(",", *it);
		if (radNS.parse(mapInfo[0]) && radEW.parse(mapInfo[1])) {
			// mapѹ
			latid = _latitude.find(radNS.toLong())->second;
			longid = _longitude.find(radEW.toLong())->second;

			// եѥϰ
			if (mapListFile->at(_baseLength + _maxFileLen * (_latitude.size()
					* longid + latid))) {
				QTextStream ts(mapListFile);
				ts << (path + *it);
				// cout << "write " << _baseLength + _maxFileLen * (_latitude.size() * longid + latid) << ":" << (path + *it) << endl;
			}
		}
		++it;
	}

	// ֥ǥ쥯ȥ⸡
	QStringList dirs = dir.entryList(QDir::Dirs);
	it = dirs.begin();
	while (it != dirs.end()) {
		if (*it != "." && *it != "..") {
			updateMapList(path + *it + "/", mapListFile);
		}
		++it;
	}
}

QString CustomMapList::searchMapFileName(long ns, long ew, Radian *radMapEW,
		Radian *radMapNS, Radian *radMapWidth, Radian *radMapHeight) {
	char tempString[512];

	map<long, int>::iterator itLat = _latitude.lower_bound(ns);
	if (itLat != _latitude.begin()) {
		map<long, int>::iterator itLatBefore = itLat;
		--itLatBefore;
		if ((ns - itLatBefore->first) <= (itLat->first - ns) || itLat
				== _latitude.end())
			itLat = itLatBefore;
	}
	_latid = itLat->second;

	map<long, int>::iterator itLong = _longitude.lower_bound(ew);
	if (itLong != _longitude.begin()) {
		map<long, int>::iterator itLongBefore = itLong;
		--itLongBefore;
		if ((ew - itLongBefore->first) <= (itLong->first - ew) || itLong
				== _longitude.end())
			itLong = itLongBefore;
	}
	_longid = itLong->second;

	_mapListFile.at(_baseLength + _maxFileLen * (_latitude.size() * _longid
			+ _latid));
	QTextStream ts(&_mapListFile);
	ts.readRawBytes(tempString, _maxFileLen);
	if (tempString[0] == '\0')
		return "";

	if (!getMapInfo(tempString, radMapEW, radMapNS, radMapWidth, radMapHeight)) {
		return "";
	}
	//cout << "ns=" << ns << ", ew=" << ew << endl;
	//cout << "latid=" << _latid << ", longid=" << _longid << endl;
	//cout << _basePath << tempString << endl;
	return _basePath + tempString;
}

QString CustomMapList::pointMapFileName(int diffLat, int diffLong,
		Radian *radMapEW, Radian *radMapNS, Radian *radMapWidth,
		Radian *radMapHeight) {
	char tempString[512];
	int latid, longid;

	latid = _latid + diffLat;
	longid = _longid + diffLong;

	if (latid < 0 || latid >= _latitude.size())
		return "";

	if (longid < 0 || longid >= _longitude.size())
		return "";

	_mapListFile.at(_baseLength + _maxFileLen * (_latitude.size() * longid
			+ latid));
	QTextStream ts(&_mapListFile);
	ts.readRawBytes(tempString, _maxFileLen);
	if (tempString[0] == '\0')
		return "";

	if (!getMapInfo(tempString, radMapEW, radMapNS, radMapWidth, radMapHeight)) {
		return "";
	}
	return _basePath + tempString;
}

bool CustomMapList::getMapInfo(QString mapFileName, Radian *radMapEW,
		Radian *radMapNS, Radian *radMapWidth, Radian *radMapHeight) {
	if (mapFileName.isEmpty())
		return false;

	int slashPoint = mapFileName.findRev('/');
	int dotPoint = mapFileName.findRev('.');
	QString tempStr = mapFileName.mid(slashPoint +1, dotPoint - slashPoint - 1);

	QStringList mapInfo = QStringList::split(",", tempStr);
	if (mapInfo.count() < 4)
		return false;

	radMapNS->parse(mapInfo[0]);
	radMapEW->parse(mapInfo[1]);
	if (mapInfo[2].find('.') == -1)
		mapInfo[2] += ".0";
	radMapHeight->parse("S0.0." + mapInfo[2]);
	if (mapInfo[3].find('.') == -1)
		mapInfo[3] += ".0";
	radMapWidth->parse("E0.0." + mapInfo[3]);

	if (radMapWidth->toLong() == 0 || radMapHeight->toLong() == 0)
		return false;

	return true;
}
