/*
 * Copyright (C) 2009 HB1 <http://www.h7.dion.ne.jp/~rasiel/zega.html>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include"drawinfo.h"
#define _DEBUG_PRINT
#ifdef _DEBUG_PRINT
#include<iostream.h>
#endif

FormInfo::FormInfo() {
	pfunc = &DrawInfo::dummy;
}

FormInfo::FormInfo(DrawInfoFunc apfunc, QStringList adata) {
	pfunc = apfunc;
	data = adata;
}

DrawInfo::DrawInfo(QPaintDevice *paintDevice, int pd_width,
		int pd_height, Route *route, GpsLogHandler *gpsLogHandler,
		Coordinate *centerCoordinate) {
	_paintDevice = paintDevice;
	_pd_width = pd_width;
	_pd_height = pd_height;
	_route = route;
	_gpsLogHandler = gpsLogHandler;
	_centerCoordinate = centerCoordinate;
	_painter = new QPainter();
	_drawBuffer = new QPixmap();
	_base_x = 100;
	_base_y = 100;
	_width = 150;
	_height = 100;
	_padding = 2;

	// ܥ
	_funcmap[QString("<font")] = &DrawInfo::setFont;
	_funcmap[QString("<color")] = &DrawInfo::setColor;
	_funcmap[QString("<br")] = &DrawInfo::newLine;
	// GPS
	_funcmap[QString("<lat")] = &DrawInfo::setLat;
	_funcmap[QString("<lon")] = &DrawInfo::setLon;
	_funcmap[QString("<date")] = &DrawInfo::setDate;
	_funcmap[QString("<time")] = &DrawInfo::setTime;
	_funcmap[QString("<speed")] = &DrawInfo::setSpeed;
	_funcmap[QString("<height")] = &DrawInfo::setHeight;
	_funcmap[QString("<direction")] = &DrawInfo::setDirection;
	_funcmap[QString("<satellite_num")] = &DrawInfo::setSatelliteNum;
	// 롼ȾϢ
	_funcmap[QString("<route_total")] = &DrawInfo::setRouteTotal;
	_funcmap[QString("<route_remain")] = &DrawInfo::setRouteRemain;
	_funcmap[QString("<route_next")] = &DrawInfo::setRouteNext;
	_funcmap[QString("<route_begin")] = &DrawInfo::setRouteBegin;
	_funcmap[QString("<route_end")] = &DrawInfo::setRouteEnd;
}

DrawInfo::~DrawInfo() {
	delete _painter;
}

void DrawInfo::setFormat(QString format) {
	FormInfo form;

	// եޥåȽ
	_forms.clear();

	// 1ܤι⤵
	form.pfunc = &DrawInfo::setLineHight;
	_forms.push_back(form);

	// եޥåȻparse
	while (format.length() > 0) {
		form.data.clear();
		int index = format.find('<');
		if (index == 0) {
			int endIndex = format.find('>');
			QString tag = format.left(endIndex);
			QStringList tagData = QStringList::split(' ', tag);
			map<QString, DrawInfoFunc>::iterator it = _funcmap.find(tagData[0]);
			if (it != _funcmap.end()) {
				form.pfunc = it->second;
				tagData.remove(tagData.begin());
				form.data = tagData;
				_forms.push_back(form);
			}

			// parseλʬ
			format.remove(0, endIndex +1);
		} else {
			// ʸ
			if (index == -1) {
				index = format.length();
			}
			form.data << format.left(index);
			form.data[0].replace(QRegExp("&[Ll][Tt];"), QString("<"));
			form.data[0].replace(QRegExp("&[Gg][Tt];"), QString(">"));
			form.pfunc = &DrawInfo::setText;
			_forms.push_back(form);

			// parseλʬ
			format.remove(0, index);
		}
	}

	// Ԥι⤵
	QStringList *height_data = &(_forms[0].data);
	int line_height = 0;
	int font_size = 12;
	_height = _padding;
	for (vector<FormInfo>::iterator it = _forms.begin(); it != _forms.end(); ++it) {
		if (it->pfunc == &DrawInfo::setFont) {
			font_size = it->data[0].toInt();
			if (line_height < font_size + _padding)
				line_height = font_size + _padding;
		} else if (it->pfunc == &DrawInfo::newLine) {
			(*height_data) << QString::number(line_height);
			height_data = &(it->data);
			_height += line_height;
			if (it +1 != _forms.end())
				if ((it +1)->pfunc == &DrawInfo::setFont) {
					line_height = 0;
				} else {
					line_height = font_size;
				}
		}
	}
	(*height_data) << QString::number(line_height);
	_height += line_height;
	_drawBuffer->resize(_width, _height);
}

void DrawInfo::setGeometry(int x, int y, int width) {
	_base_x = x;
	_base_y = y;
	_width = width;
	_drawBuffer->resize(_width, _height);
}

void DrawInfo::resize(int width, int height) {
	_pd_width = width;
	_pd_height = height;
}

void DrawInfo::repaint() {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::repaint()" << endl;
#endif
	if (_gpsLogHandler->isAlive())
		_coordinate = _gpsLogHandler->getNowCoordinate();

	_line_height = 0;
	_x = 0;
	_y = 0;
	_painter->begin(_drawBuffer);
	_painter->setPen(QPen(Qt::black, 1, Qt::SolidLine));
	_painter->setBrush(QBrush(Qt::white, Qt::SolidPattern));
	_painter->drawRect(0, 0, _width, _height);
	_x += _padding;
	_strData = "";
	for (vector<FormInfo>::iterator it = _forms.begin(); it != _forms.end(); ++it) {
		(this->*(it->pfunc))(&(it->data));
	}
	drawText();
	_painter->end();
	
	int base_x = _base_x;
	if (base_x < 0)
		base_x = _pd_width + _base_x - _width;
	int base_y = _base_y;
	if (base_y < 0)
		base_y = _pd_height + _base_y - _height;

	bitBlt(_paintDevice, base_x, base_y,
	       _drawBuffer, 0, 0, _width, _height,
	       Qt::CopyROP, false);
}

void DrawInfo::dummy(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::dummy()" << endl;
#endif
}

void DrawInfo::setLineHight(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setHight()" << endl;
#endif
	_line_height = (*pstrlist)[0].toInt();
	_painter->setFont(QFont("lcfont", 12));
}

void DrawInfo::setFont(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setFont()" << endl;
#endif
	drawText();
	int font_size = (*pstrlist)[0].toInt();
	if (font_size > 0)
		if (pstrlist->count() < 2)
			_painter->setFont(QFont("lcfont", font_size));
		else
			_painter->setFont(QFont((*pstrlist)[1], font_size));
}

void DrawInfo::setColor(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setColor()" << endl;
#endif
	if (pstrlist->count() < 1) {
#ifdef _DEBUG_PRINT
		cout << "not set parameter" << endl;
#endif
		return;
	}

	drawText();
	QColor fontColor((*pstrlist)[0]);
	_painter->setPen(fontColor);
}

void DrawInfo::setText(const QStringList *pstrlist) {
	_strData += (*pstrlist)[0];
}

void DrawInfo::drawText() {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::drawText()" << endl;
#endif
	if (_strData.length() > 0) {
		QRect drawRect;
		_painter->drawText(_x, _y, _width - _padding *2, _line_height,
				Qt::AlignLeft | Qt::AlignVCenter, _strData, -1, &drawRect);
		_x += drawRect.width();
		_strData = "";
	}
}

void DrawInfo::newLine(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::newLine() " << (*pstrlist)[0] << endl;
#endif
	drawText();
	_x = _padding;
	_y += _line_height;
	_line_height = (*pstrlist)[0].toInt();
}

void DrawInfo::setLat(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setLat()" << endl;
#endif
	_strData += _coordinate.getRadNS().toNS();
}

void DrawInfo::setLon(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setLon()" << endl;
#endif
	_strData += _coordinate.getRadEW().toEW();
}

void DrawInfo::setDate(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setDate()" << endl;
#endif
	int addSecs = 0;
	if(pstrlist->count() >= 1)
		addSecs = (*pstrlist)[0].toInt();
	QDateTime drawTime = _gpsLogHandler->getNowTime();
	drawTime = drawTime.addSecs(addSecs * 60);
	QDate drawDate = drawTime.date();
	_strData += QString::number(drawDate.year()) + "-" + QString::number(drawDate.month()) + "-" + QString::number(drawDate.day()) +  " " + drawTime.time().toString();
}

void DrawInfo::setTime(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setTime()" << endl;
#endif
	int addSecs = 0;
	if(pstrlist->count() >= 1)
		addSecs = (*pstrlist)[0].toInt();
	QDateTime drawTime = _gpsLogHandler->getNowTime();
	drawTime = drawTime.addSecs(addSecs * 60);
	_strData += drawTime.time().toString();
}

void DrawInfo::setSpeed(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setSpeed()" << endl;
#endif
	_strData += QString::number(_gpsLogHandler->getNowSpeed(), 'f', 2) + "km";
}

void DrawInfo::setHeight(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setHeight()" << endl;
#endif
	_strData += QString::number(_coordinate.getIntHight() * 0.01, 'f', 2) + "m";
}

void DrawInfo::setDirection(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setDirection()" << endl;
#endif
	_strData += QString::number(_gpsLogHandler->getNowDirection(), 'f', 2);
}

void DrawInfo::setSatelliteNum(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setSatelliteNum()" << endl;
#endif
	_strData += QString::number(_gpsLogHandler->getSatelliteNum());
}

void DrawInfo::setRouteTotal(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setRouteTotal()" << endl;
#endif
	_strData += QString::number(_route->_routeLength, 'f', 3) + "km";
}

void DrawInfo::setRouteRemain(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setRouteRemain()" << endl;
#endif
	if(_route->_route.empty())
		_strData += "none";
	else
		_strData += QString::number(_route->getRemain(), 'f', 3) + "km";
}

void DrawInfo::setRouteNext(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setRouteNext()" << endl;
#endif
	if(_route->_route.empty())
		_strData += "none";
	else
		_strData += QString::number(_route->_currentNearLength *1000, 'f', 0) + "m";
}

void DrawInfo::setRouteBegin(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setRouteBegin()" << endl;
#endif
	_strData += QString::number(_route->getLength(_route->_route.begin(), _centerCoordinate), 'f', 3) + "km";
}

void DrawInfo::setRouteEnd(const QStringList *pstrlist) {
#ifdef _DEBUG_PRINT
	cout << "call DrawInfo::setRouteEnd()" << endl;
#endif
	_strData += QString::number(_route->getLength(&(_route->_route.back()), _centerCoordinate), 'f', 3) + "km";
}
