/*
 * Copyright (C) 2009 HB1 <http://www.h7.dion.ne.jp/~rasiel/zega.html>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef GPSLOGHANDLER_H
#define GPSLOGHANDLER_H

#include<qpe/qpeapplication.h>
#include<pthread.h>
#include<qdir.h>
#include<qfile.h>
#include<qtextstream.h>
#include<qstring.h>
#include<qstringlist.h>
#include<qdatetime.h>
#include<vector.h>
#include"coordinate.h"

#include<iostream.h>

#define KM_PER_KNOT 1.852

extern QPEApplication *_app;

class GpsLogHandler {
public:
	GpsLogHandler(int recentTerm = 1, int recentLogNum = 600,
			int obsolateTerm = 10);
	~GpsLogHandler();

	void start(void);
	void run(void);
	void stop(void);
	inline bool isAlive(void) {
		return _alive;
	};
	Coordinate getNowCoordinate();
	double getNowDirection();
	double getNowSpeed();
	int getSatelliteNum();
	QDateTime getNowTime();
	void setLogDir(QString logDir);
	QString getLogDir();
	void setDevFile(QString devFile);
	void setGpsOutput(QString gpsOutput);
	void setStartGpsCommand(QString cmd);
	void setStopGpsCommand(QString cmd);
	void loadObsolateLogFile(QString logFile);
	bool isStopped();
	void changeDebugPrint();
	void setRecentLogFirst();
	bool endRecentLog();
	Coordinate getNextRecentLog();
	void setObsolateLogFirst();
	bool endObsolateLog();
	Coordinate getNextObsolateLog();
	void setGuessLowLevelFixed(bool flag);

private:
	QString _basePath;
	QString _gpsLogDir;
	QString _startGpsCommand;
	QString _stopGpsCommand;
	QFile _gpsLogFile;
	QFile _gpsDevFile;
	int _satelliteNum;
	volatile bool _stopped;
	volatile bool _alive;
	bool _debug_print;
	bool _gpsOutputWGS84;
	bool _guessLowLevelFixed;
	Coordinate _coordinate;
	double _direction;
	double _speed;
	QDateTime _nowTime;
	pthread_t _thread_id;
	pthread_mutex_t _mutex_id;
	int _recentTerm;
	int _recentLogNum;
	int _recentNext;
	int _recentPtr;
	bool _recentLogFull;
	Coordinate* _recentLog;
	int _obsolateTerm;
	vector<Coordinate>::iterator _obsolateIt;
	vector<Coordinate> _obsolateLog;
};

void* thread_func(void *pGpsLogHandler);

#endif //GPSLOGHANDLER_H
