/*
 * Copyright (C) 2009 HB1 <http://www.h7.dion.ne.jp/~rasiel/zega.html>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include<algorithm>
#include<math.h>
#include<qfile.h>
#include"route.h"

Route::Route(){
	clear();
}

// ֤εΥ򻻽
double Route::getLength(Coordinate *aPoint, Coordinate *bPoint) {
	double x, y;
	long xdeg = aPoint->getRadEW().toLong() - bPoint->getRadEW().toLong();
	double xrad = xdeg * M_PI / ( 600000.0 * 180.0 );
	x = xdeg * 10.0 * cos(xrad) / 54000.0;
	y = (aPoint->getRadNS().toLong() - bPoint->getRadNS().toLong()) * 10.0 / 54000.0;
	return sqrt(x * x + y * y);
}

// 
void Route::clear(){
	// ϩ
	_route.clear();
	
	// ƵΥ
	_routeLength = 0.0;
	_remainRouteLength = 0.0;
	_currentNearLength = 0.0;

	// ľΥݥȤꥻå
	_nearRouteIndex = -1;
}

// 롼ȵΥƷ׻
void Route::resetLength(){
	// 롼ȵΥƷ׻
	for (vector<Coordinate>::iterator it = _route.begin(); it + 1
			!= _route.end(); ++it) {
		_routeLength = _routeLength + getLength(it, it + 1);
	}
	// ľΥݥȤꥻå
	_nearRouteIndex = -1;	
}

// 롼ȥեɹ
int Route::loadFile(QString routeFileName){
	QFile routeFile;
	bool loadKml = false;
	if (routeFileName.right(4) == ".kml") {
		loadKml = true;
	}
	routeFile.setName(routeFileName);
	if (routeFile.exists()) {
		if (!routeFile.open(IO_ReadOnly)) {
			// ե뤬ʤ
			return 1;
		}
	} else {
		// ե뤬ʤ
		return 2;
	}

	char buf[512];
	if (loadKml) {
		while (!routeFile.atEnd()) {
			int sz = routeFile.readLine(buf, sizeof(buf));
			if (sz <= 0)
				break;
			if (strstr(buf, "<coordinates>") != NULL)
				break;
		}
	}

	_route.clear();
	_routeLength = 0.0;
	Radian radNS, radEW;
	while (!routeFile.atEnd()) {
		int sz = routeFile.readLine(buf, sizeof(buf));
		if (sz <= 0)
			break;
		const QStringList& qslTemp = QStringList::split(",", buf);
		int intHight = 0;
		if (loadKml && qslTemp.count() >= 3) {
			intHight = qslTemp[2].toInt();
		}
		if (qslTemp.count() >= 2) {
			radEW.parse2("E" + qslTemp[0]);
			radNS.parse2("N" + qslTemp[1]);
			Coordinate coordinate(radEW, radNS, intHight);
			if (loadKml)
				coordinate.convWGS84ToTokyo();
			_route.push_back(coordinate);
		}
	}
	routeFile.close();
	
	resetLength();
	
	return 0;
}

// 롼Ȼķ׻
bool Route::setRemain(Coordinate coordinate){
	if(_route.size() == 0) {
		return false;
	}

	long lns = coordinate.getRadNS().toLong();
	long lew = coordinate.getRadEW().toLong();

	if(_nearRouteIndex == -1) {
		long distMin = -1;
		long dist;
		for(int i = 0; i < _route.size(); i++) {
			dist = abs(_route[i].getRadNS().toLong() - lns)
			       + abs(_route[i].getRadEW().toLong() - lew);
			if(dist < distMin || distMin == -1) {
				distMin = dist;
				_nearRouteIndex = i;
				_remainRouteLength = 0;
			} else {
				_remainRouteLength += getLength(&(_route[i]), &(_route[i-1]));
			}
		}
		_remainNearNext = false;
	} else {
		long distMin = -1;
		long dist;
		distMin = abs(_route[_nearRouteIndex].getRadNS().toLong() - lns)
		+ abs(_route[_nearRouteIndex].getRadEW().toLong() - lew);
		for(int i = _nearRouteIndex +1; i < _route.size(); i++) {
			dist = abs(_route[i].getRadNS().toLong() - lns)
			+ abs(_route[i].getRadEW().toLong() - lew);
			if (dist < distMin) {
				distMin = dist;
				_nearRouteIndex = i;
				if (_remainNearNext) {
					_remainNearNext = false;
				} else {
					_remainRouteLength -= getLength(&(_route[i]), &(_route[i-1]));
				}
			} else {
				break;
			}
		}
	}
	if (_nearRouteIndex + 1 < _route.size() &&
			((_route[_nearRouteIndex].getRadNS().toLong() < lns && lns < _route[_nearRouteIndex +1].getRadNS().toLong()) ||
					(_route[_nearRouteIndex].getRadNS().toLong() > lns && lns > _route[_nearRouteIndex +1].getRadNS().toLong()) ||
					(_route[_nearRouteIndex].getRadEW().toLong() < lew && lew < _route[_nearRouteIndex +1].getRadEW().toLong()) ||
					(_route[_nearRouteIndex].getRadEW().toLong() > lew && lew > _route[_nearRouteIndex +1].getRadEW().toLong())) ) {
		_currentNearLength = getLength(&coordinate, &(_route[_nearRouteIndex +1]));
		if (!_remainNearNext) {
			_remainNearNext = true;
			_remainRouteLength -= getLength(&(_route[_nearRouteIndex]), &(_route[_nearRouteIndex +1]));
		}
	} else {
		_currentNearLength = getLength(&coordinate, &(_route[_nearRouteIndex]));
		if (_remainNearNext) {
			_remainNearNext = false;
			_remainRouteLength += getLength(&(_route[_nearRouteIndex]), &(_route[_nearRouteIndex +1]));
		}
	}

	return true;
}

// 롼Ȼļ
double Route::getRemain(){
	return _remainRouteLength + _currentNearLength;
}

// 롼ȤκǸɲ
void Route::addPoint(Coordinate coordinate){
	// 롼Υ
	if (_route.size() >= 1)
		_routeLength = _routeLength + getLength(&(_route[_route.size() -1]), &coordinate);

	// 롼Ȥ˥ݥȤɲ
	_route.push_back(coordinate);

	// ľΥݥȤꥻå
	_nearRouteIndex = -1;	
}

// 롼ȺǸΥݥȤ
void Route::deleteLastPoint(){
	if (_route.size() <= 0)
		return;
	
	// ݥȤ2İʾ¸ߤΥޥʥ
	if (_route.size() >= 2)
		_routeLength = _routeLength - getLength(&(_route[_route.size() -1]), &(_route[_route.size() -2]));
	else
		_routeLength = 0.0;
	
	// ǸΥݥȤ
	_route.pop_back();

	// ľΥݥȤꥻå
	_nearRouteIndex = -1;	
}

// 롼ȿž
void Route::reverse(){
    // ݥȤո¤ؤ
	std::reverse(_route.begin(), _route.end());

	// ľΥݥȤꥻå
	_nearRouteIndex = -1;
}
