/*
 * Copyright (C) 2009 HB1 <http://www.h7.dion.ne.jp/~rasiel/zega.html>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include"zega.h"
#include<algorithm>
#include<math.h>
#include<qpe/qcopenvelope_qws.h>
//#define _DEBUG_PRINT

QPoint rotate(QPoint point, double deg) {
	QPoint rtnPoint;
	double rad = deg * M_PI / 180.0;
	rtnPoint.setX(point.x() * cos(rad) * -1 + point.y() * sin(rad));
	rtnPoint.setY((point.x() * sin(rad) + point.y() * cos(rad)) * -1);
	return rtnPoint;
}

Config Zega::config = Config("Zega");

Zega::Zega(QWidget* parent, const char* name) :
	QWidget(parent, name) {
	// ɤ߹
	config.setGroup("Color");
	_routeColor.setNamedColor("#FF0000");
	if (!config.hasKey("routeColor"))
		config.writeEntry("routeColor", "#FF0000");
	else
		_routeColor.setNamedColor(config.readEntry("routeColor"));

	_recentLogColor.setNamedColor("#FFFF00");
	if (!config.hasKey("recentLogColor"))
		config.writeEntry("recentLogColor", "#FFFF00");
	else
		_recentLogColor.setNamedColor(config.readEntry("recentLogColor"));

	_obsoluteLogColor.setNamedColor("#00FFFF");
	if (!config.hasKey("obsoluteLogColor"))
		config.writeEntry("obsoluteLogColor", "#00FFFF");
	else
		_obsoluteLogColor.setNamedColor(config.readEntry("obsoluteLogColor"));

	_currentPointColor.setNamedColor("#00FF00");
	if (!config.hasKey("currentPointColor"))
		config.writeEntry("currentPointColor", "#00FF00");
	else
		_currentPointColor.setNamedColor(config.readEntry("currentPointColor"));

	_destinationLineColor.setNamedColor("#EF810F");
	if (!config.hasKey("destinationLineColor"))
		config.writeEntry("destinationLineColor", "#EF810F");
	else
		_destinationLineColor.setNamedColor(config.readEntry("destinationLineColor"));

	// 
	config.setGroup("Width");
	_routeWidth = 6;
	if (!config.hasKey("routeWidth"))
		config.writeEntry("routeWidth", 6);
	else
		_routeWidth = config.readNumEntry("routeWidth");
	if (_routeWidth < 1)
		_routeWidth = 1;

	_logWidth = 6;
	if (!config.hasKey("logWidth"))
		config.writeEntry("logWidth", 6);
	else
		_logWidth = config.readNumEntry("logWidth");
	if (_logWidth < 1)
		_logWidth = 1;

	_destinationLineWidth = 2;
	if (!config.hasKey("destinationLineWidth"))
		config.writeEntry("destinationLineWidth", 2);
	else
		_destinationLineWidth = config.readNumEntry("destinationLineWidth");

	if (_destinationLineWidth < 1)
		_destinationLineWidth = 1;

	// 륭ˤϿްưΥ
	config.setGroup("Keys");
	_cursorMoveDen = 2;
	if (!config.hasKey("cursorMoveDen"))
		config.writeEntry("cursorMoveDen", 2);
	else
		_cursorMoveDen = config.readNumEntry("cursorMoveDen");
	if (_cursorMoveDen < 1)
		_cursorMoveDen = 1;

	_shiftCursorMoveDen = 1;
	if (!config.hasKey("shiftCursorMoveDen"))
		config.writeEntry("shiftCursorMoveDen", 1);
	else
		_shiftCursorMoveDen = config.readNumEntry("shiftCursorMoveDen");
	if (_shiftCursorMoveDen < 1)
		_shiftCursorMoveDen = 1;

	// ɽե饰
	_autoCentered = true;
	_showGpsLog = true;
	_showRoute = true;
	_showDestinationLine = true;

	// GPS楫
	_evenRePaint = true;

	// ɥȥ륿ȥ
	setCaption(tr("Zega"));

	// ˥塼Сɲ
	_menu = new QMenuBar(this, "menubar");

	_filemenu = new QPopupMenu(this, "filemenu");
	_filemenu->insertItem(tr("&Load File"), this, SLOT(slotLoad()));
	_filemenu->insertItem(tr("&Save Route File"), this, SLOT(slotSaveRoute()));
	_filemenu->insertSeparator();
	_filemenu->insertItem(tr("&Quit"), this, SLOT(close()));

	// Ͽޥե륻åȾɤ߹
	config.setGroup("Maps");

	// ХޥåNavi̵Ͻ
	_mobileMapNaviMaps = NULL;
	if ( !config.hasKey("mobileMapNaviMapDir") ) {
		config.writeEntry("mobileMapNaviMapDir", "/mnt/card/sd_map");

		if ( !config.hasKey("mapNum") )
			config.writeEntry("mapNum", 0);
		config.writeEntry("mapNum", 11);
		config.writeEntry("map1", tr("ХޥåNavi,M,0"));
		config.writeEntry("map2", tr("ХޥåNaviϿ,M,1"));
		config.writeEntry("map3", tr("ХޥåNaviϿ,M,2"));
		config.writeEntry("map4", tr("ХޥåNaviϿ,M,3"));
		config.writeEntry("map5", tr("ХޥåNaviܺϿ,M,4"));
		config.writeEntry("map6", tr("GM_LiteϿ6,G,6"));
		config.writeEntry("map7", tr("GM_LiteϿ7,G,7"));
		config.writeEntry("map8", tr("GM_LiteϿ8,G,8"));
		config.writeEntry("map9", tr("GM_LiteϿ9,G,9"));
		config.writeEntry("map10", tr("GM_LiteϿ10,G,10"));
		config.writeEntry("map11", tr("ޥå 1/150000,C,/mnt/card/maps/1"));
	}
	QDir mmnDir(config.readEntry("mobileMapNaviMapDir"));
	if (mmnDir.exists()) {
		_mobileMapNaviMaps = new MobileMapNaviMaps(config.readEntry("mobileMapNaviMapDir"));
	}

	// GM_Liteޥå׽
	_GM_LiteMaps = NULL;
	if ( !config.hasKey("GM_LiteMapDir") ) {
		config.writeEntry("GM_LiteMapDir", "/mnt/card/gm_lite/imgs");
	}
	QDir gmlDir(config.readEntry("GM_LiteMapDir"));
	if (gmlDir.exists()) {
		_GM_LiteMaps = new GM_LiteMaps(config.readEntry("GM_LiteMapDir"));
	}

	// Ͽ˥塼ɲ
	_scalemenu = new QPopupMenu(this, "scalemenu");
	int mapNum = config.readNumEntry("mapNum");
	QString strTmp;
	QString strNumTmp;
	int intMapId = 0;
	mapDirs.clear();
	for (int i = 1; i <= mapNum; i++) {
		bool addScaleMenu = true;
		strTmp = QString("map") + strNumTmp.setNum(i);
		if (config.hasKey(strTmp) ) {
			QStringList mapInfo = config.readListEntry(strTmp, ',');
			if (mapInfo.count() >= 3) {
				if (_mobileMapNaviMaps == NULL && mapInfo[1] == "M")
					addScaleMenu = false;
				if (_GM_LiteMaps == NULL && mapInfo[1] == "G")
					addScaleMenu = false;

				if (addScaleMenu) {
					mapDirs.append(config.readEntry(strTmp));
					int itemId = _scalemenu->insertItem(mapInfo[0], this, SLOT(slotChScale(int)));
					_scalemenu->setItemParameter(itemId, intMapId);
					intMapId++;
				}
			}
		}
	}

	_gpsmenu = new QPopupMenu(this, "gpsmenu");
	_gpsmenu->insertItem(tr("&Start GPS"), this, SLOT(startGps()));
	_gpsmenu->insertItem(tr("S&top GPS"), this, SLOT(stopGps()));

	_editmenu = new QPopupMenu(this, "editmenu");
	_editmenu->insertItem(tr("&Add Route Point"), this,
			SLOT(slotAddRoutePoint()));
	_editmenu->insertItem(tr("&Delete Route Point"), this,
			SLOT(slotDelRoutePoint()));
	_editmenu->insertItem(tr("&Reverse Route"), this, SLOT(slotReverseRoute()));
	_editmenu->insertItem(tr("&Clear Route"), this, SLOT(slotClearRoute()));
	_editmenu->insertItem(tr("Calc Remain Length"), this, SLOT(slotCalcRemainLength()));
	
	_displaymenu = new QPopupMenu(this, "displaymenu");
	_displaymenu->insertItem(tr("&GPS Log On/Off"), this,
			SLOT(slotToggleShowGpsLog()));
	_displaymenu->insertItem(tr("&Route On/Off"), this,
			SLOT(slotToggleShowRoute()));
	_displaymenu->insertItem(tr("&Auto Centered On/Off"), this,
			SLOT(slotToggleAutoCentered()));

	// إѤΥ٥
	_help = new Help(this);

	_helpmenu = new QPopupMenu(this, "helpmenu");
	_helpmenu->insertItem(tr("&Help"), this, SLOT(slotHelp()));
	_helpmenu->insertItem(tr("&About"), this, SLOT(about()));

	_menu->insertItem(tr("&File"), _filemenu);
	_menu->insertItem(tr("&Scale"), _scalemenu);
	_menu->insertItem(tr("&GPS"), _gpsmenu);
	_menu->insertItem(tr("&Edit"), _editmenu);
	_menu->insertItem(tr("&Display"), _displaymenu);
	_menu->insertItem(tr("&Help"), _helpmenu);

	config.setGroup("CurrentCoordinate");
	// ֤߰¸ߤʤؤ
	if (!config.hasKey("lat"))
		config.writeEntry("lat", "N35.40.40.00");
	if (!config.hasKey("lon"))
		config.writeEntry("lon", "E139.46.15.00");

	Radian tempRad;
	tempRad.parse(config.readEntry("lat"));
	_center.setRadNS(tempRad);
	tempRad.parse(config.readEntry("lon"));
	_center.setRadEW(tempRad);
	_pixelNS = 0;
	_pixelEW = 0;
	_pixelBase = PIXEL_BASE;

	// Ͽޤ
	_mapIndex = 0;
	config.setGroup("Maps");
	if (!config.hasKey("currentMap"))
		config.writeEntry("currentMap", 0);
	else {
		_mapIndex = config.readNumEntry("currentMap");
		if (_mapIndex >= mapDirs.count())
			_mapIndex = 0;
	}

	// ޥåץå
	if (!config.hasKey("mapCacheSize"))
		config.writeEntry("mapCacheSize", 12);
	int mapCacheSize = config.readNumEntry("mapCacheSize");
	_mapCache = new MapCache(mapCacheSize);

	_mapList = &_customMapList;
	_buffer = NULL;
	_mapBuffer = NULL;
	slotChScale(_mapIndex);

	// GPS
	config.setGroup("GPS");
	// GPSեѥ
	if ( !config.hasKey("logDir") )
		config.writeEntry("logDir", "/mnt/card/gpslog");
	// GPSǥХե
	if ( !config.hasKey("devFile") )
		config.writeEntry("devFile", "/dev/ttyS3");
	// NMEA-0183Ϥ¬Ϸ
	if ( !config.hasKey("gpsOutput") )
		config.writeEntry("gpsOutput", "WGS84");
	// GPSϥޥ
	if ( !config.hasKey("startGpsCommand") )
		config.writeEntry("startGpsCommand",
				QString("cardctl resume 0;stty 4800 < ")
						+ config.readEntry("devFile"));
	// GPSλޥ
	if ( !config.hasKey("stopGpsCommand") )
		config.writeEntry("stopGpsCommand", "cardctl suspend 0");

	// GPS¬̾ª3İʾޤ
	if ( !config.hasKey("guessLowLevelFixed") )
		config.writeEntry("guessLowLevelFixed", false);
	
	// ľȤGPS־򵭲ֳ
	if ( !config.hasKey("recentTerm") )
		config.writeEntry("recentTerm", "1");
	int recentTerm = config.readNumEntry("recentTerm");
	if (recentTerm <= 0)
		recentTerm = 1;

	// ľȤGPS־Ͽ
	if ( !config.hasKey("recentLogNum") )
		config.writeEntry("recentLogNum", "600");
	int recentLogNum = config.readNumEntry("recentLogNum");
	if (recentLogNum <= 0)
		recentLogNum = 600;

	// ŤȤGPS־򵭲ֳ
	if ( !config.hasKey("obsolateTerm") )
		config.writeEntry("obsolateTerm", "10");
	int obsolateTerm = config.readNumEntry("obsolateTerm");
	if (obsolateTerm <= recentTerm)
		obsolateTerm = recentTerm * 4;

	_gpsLogHandler = new GpsLogHandler(recentTerm, recentLogNum, obsolateTerm);
	_gpsLogHandler->setLogDir(config.readEntry("logDir"));
	_gpsLogHandler->setDevFile(config.readEntry("devFile"));
	_gpsLogHandler->setGpsOutput(config.readEntry("gpsOutput"));
	_gpsLogHandler->setStartGpsCommand(config.readEntry("startGpsCommand"));
	_gpsLogHandler->setStopGpsCommand(config.readEntry("stopGpsCommand"));
	_gpsLogHandler->setGuessLowLevelFixed(config.readBoolEntry("guessLowLevelFixed"));

	// ưǥ󥿥󥰤볰Ȥ
	if ( !config.hasKey("autoCenteredBounds") )
		config.writeEntry("autoCenteredBounds", 50);
	_autoCenteredBounds = config.readNumEntry("autoCenteredBounds");
	if (_autoCenteredBounds <= 0)
		_autoCenteredBounds = 0;

	_mapBuffer = new QPixmap();

	// ɽ
	config.setGroup("InfoView");
	if (!config.hasKey("GpsInfoFormat"))
		config.writeEntry("GpsInfoFormat", "<font 12><color #0000FF>GPS Info<br>"
					"<font 16><color #000000><time 540><font 12> (<satellite_num>)<br>"
					"lat: <lat><br>"
					"lon: <lon><br>"
					"direction: <direction><br>"
					"speed: <font 18><speed><br>"
					"<font 12>height: <height><br>"
					"remain: <font 18><color #FF0000><route_remain><br>"
					"<font 12><color #000000>next: <font 16><route_next>"
				);
	if (!config.hasKey("GpsInfoX"))
		config.writeEntry("GpsInfoX", -5);
	if (!config.hasKey("GpsInfoY"))
		config.writeEntry("GpsInfoY", 32);
	if (!config.hasKey("GpsInfoWidth"))
		config.writeEntry("GpsInfoWidth", 120);

	_drawGpsInfo = new DrawInfo(this, this->width(), this->height(), &_route, _gpsLogHandler, &_center);
	_drawGpsInfo->setFormat(config.readEntry("GpsInfoFormat"));
	_drawGpsInfo->setGeometry(config.readNumEntry("GpsInfoX"), config.readNumEntry("GpsInfoY"), config.readNumEntry("GpsInfoWidth"));

	if (!config.hasKey("RouteInfoFormat"))
		config.writeEntry("RouteInfoFormat", "<font 12><color #0000FF>Route Info<br>"
					"<color #000000>total: <route_total><br>"
					"begin: <route_begin><br>"
					"end: <route_end>"
				);
	if (!config.hasKey("RouteInfoX"))
		config.writeEntry("RouteInfoX", 5);
	if (!config.hasKey("RouteInfoY"))
		config.writeEntry("RouteInfoY", -5);
	if (!config.hasKey("RouteInfoWidth"))
		config.writeEntry("RouteInfoWidth", 100);

	_drawMapInfo = new DrawInfo(_mapBuffer, _mapBuffer->width(), _mapBuffer->height(), &_route, _gpsLogHandler, &_center);
	_drawMapInfo->setFormat(config.readEntry("RouteInfoFormat"));
	_drawMapInfo->setGeometry(config.readNumEntry("RouteInfoX"), config.readNumEntry("RouteInfoY"), config.readNumEntry("RouteInfoWidth"));

	// ̹ޡ
	_updateTimer = new QTimer( this );
	connect(_updateTimer, SIGNAL(timeout()), this, SLOT(slotRePaint()) );
	
	setWFlags(WStyle_Customize | WStyle_NoBorder);
}

// Zegaλ
Zega::~Zega() {
	if (_gpsLogHandler->isAlive())
		stopGps();

	config.setGroup("CurrentCoordinate");
	config.writeEntry("lat", _center.getRadNS().toNS());
	config.writeEntry("lon", _center.getRadEW().toEW());
	config.setGroup("Maps");
	config.writeEntry("currentMap", (int)_mapIndex);
}

// ڥ̵
void Zega::disableSuspend() {
	config.setGroup("General");
	if ( !config.hasKey("controlZaurusSuspend") )
		config.writeEntry("controlZaurusSuspend", "true");

	if (config.readEntry("controlZaurusSuspend") != "true")
		return;

	// ưȼưڥɤ̵ˤ
	/* +--+--+ڥ+
	 * |0 |      |          |
	 * |1 |      |    o     |
	 * |2 |  o   |          |
	 * |3 |  o   |    o     |
	 */
	QCopEnvelope("QPE/System", "setScreenSaverMode(int)") << 0;
	// ꡼󥻡Ф̵ˤ
	QCopEnvelope("QPE/System", "setScreenSaverIntervals(int,int,int)") << 0
			<< 0 << 0;
}

// ڥͭ
void Zega::enableSuspend() {
	config.setGroup("General");
	if (config.readEntry("controlZaurusSuspend") != "true")
		return;

	// ꡼󥻡Фͭˤ
	QCopEnvelope("QPE/System", "setScreenSaverIntervals(int,int,int)") << -1
			<< -1 << -1;
	// ưȥȥѥդͭˤ
	QCopEnvelope("QPE/System", "setScreenSaverMode(int)") << 3;
}

// ϿޥåȤѹ
void Zega::slotChScale(int mapId) {
	_mapIndex = mapId;
	const QStringList& qslTemp = QStringList::split(",", mapDirs[mapId]);
#ifdef _DEBUG_PRINT
	cout << qslTemp[2] << ":" << mapDirs[mapId] << endl;
#endif
	if (qslTemp[1] == "C") {
		_customMapList.changeMap(qslTemp[2]);
		_mapList = &_customMapList;
	} else if (qslTemp[1] == "G") {
		if (_GM_LiteMaps == NULL)
			return;
		_GM_LiteMaps->changeScale(qslTemp[2]);
		_mapList = _GM_LiteMaps;
	} else if (qslTemp[1] == "M") {
		if (_mobileMapNaviMaps == NULL)
			return;
		_mobileMapNaviMaps->changeScale(qslTemp[2]);
		_mapList = _mobileMapNaviMaps;
	} else {
		return;
	}

	if (_mapBuffer != NULL)
		paintMaps();
}

// ̺
void Zega::slotRePaint() {
	static int beforeX = -50;
	static int beforeY = -50;
	Coordinate coordinate;
	_evenRePaint = !_evenRePaint;

	// GPS̤¬̤ξμơɽ
	if (_gpsLogHandler->isAlive() == false) {
		QPainter painter(this);
		if (_evenRePaint) {
			painter.setPen(QPen(darkGray, 1, SolidLine));
			painter.setBrush(QBrush(darkGray, SolidPattern));
		} else {
			painter.setPen(QPen(black, 1, SolidLine));
			painter.setBrush(QBrush(black, SolidPattern));
		}
		painter.drawEllipse(10, 40, 10, 10);
		painter.end();
		_drawGpsInfo->repaint();
		return;
	}

	// ּ߰
	coordinate = _gpsLogHandler->getNowCoordinate();
	long lns = coordinate.getRadNS().toLong();
	long lew = coordinate.getRadEW().toLong();
	double direction = _gpsLogHandler->getNowDirection();
	_route.setRemain(coordinate);
	int x = ((lew - _center.getRadEW().toLong()) << _pixelBase)/ _pixelEW
			+ this->width()/2;
	int y = ((lns - _center.getRadNS().toLong()) << _pixelBase)/ _pixelNS
			+ this->height()/2;

	// ֤߰ϰϳξϸ֤߰ѹ
	if (_autoCentered) {
		if ( x < _autoCenteredBounds || x > this->width() -_autoCenteredBounds || y < _autoCenteredBounds || y > this->height() - _autoCenteredBounds) {
			_center = coordinate;
			paintMaps();
			x = this->width()/2;
			y = this->height()/2;
		}
	}

	// õ
	if (beforeX >= 0 && beforeX < this->width() && beforeY >= 0 && beforeY
			< this->height()) {
		bitBlt(this, beforeX-27, beforeY-27, _mapBuffer, beforeX-27,
				beforeY-27, 55, 55);
	}

	// ֤߰Ͽޤ˥ݥ
	if (x >= 0 && x < this->width() && y >= 0 && y < this->height()) {
		// ϰϵϿ
		beforeX = x;
		beforeY = y;

		// ֤߰ޤǤεפХåե˵Ͽ
		QPainter painter(_mapBuffer);
		painter.setPen(QPen(blue, 1, SolidLine));
		painter.setBrush(QBrush(_recentLogColor, SolidPattern));
		painter.drawEllipse(x - _logWidth/2, y -_logWidth/2, _logWidth,
				_logWidth);
		painter.end();

		// 		
		painter.begin(this);
		painter.translate(x, y);
		painter.setPen(QPen(black, 1, SolidLine));
		painter.setBrush(QBrush(_currentPointColor, SolidPattern));
		QPointArray tempPoints(3);
		tempPoints.setPoint(0, rotate(QPoint(0, 10), direction));
		tempPoints.setPoint(1, rotate(QPoint(-5, -9), direction));
		tempPoints.setPoint(2, rotate(QPoint(5, -9), direction));
		painter.drawPolygon(tempPoints);
		painter.setPen(QPen(black, 1, Qt::DashLine));
		painter.setBrush(Qt::NoBrush);
		painter.drawEllipse(-25, -25, 51, 51);
		painter.end();
	}
	QPainter painter(this);
	
	// GPS򥢥ƥ֤ɽ
	if (_evenRePaint) {
		painter.setPen(QPen(darkRed, 1, SolidLine));
		painter.setBrush(QBrush(darkRed, SolidPattern));
	} else {
		painter.setPen(QPen(red, 1, SolidLine));
		painter.setBrush(QBrush(red, SolidPattern));
	}
	painter.drawEllipse(10, 40, 10, 10);
	painter.end();

	// ־ɽ
	_drawGpsInfo->repaint();
}

void Zega::startGps() {
	config.setGroup("GPS");
	if ( !config.hasKey("UpdateInterval") ) {
		config.writeEntry("UpdateInterval", 1000);
	}
	int intInterval = config.readNumEntry("UpdateInterval");

	QDir dir;
	QString logDir = _gpsLogHandler->getLogDir();
	if (!dir.exists(logDir)) {
		int rtn = QMessageBox::warning(this, tr("GPS Log Dir Not Exist."),
				tr("GPSǥ쥯ȥ(" + logDir + ")ʤ衣\näƤ"), tr(""),
				tr(""), QString::null, 1);
		if (rtn == 0)
		QDir().mkdir(logDir);
	}
	if (!dir.exists(logDir)) {
		QMessageBox::critical(this, tr("Cannot Start GPS"),
		tr("GPSǥ쥯ȥ(" + logDir + ")ʤ\nGPS򤢤ޤ͡"));
		return;
	}

	// LCD꡼󥻡Сzaurusڥ̵
	disableSuspend();

	_gpsLogHandler->start();
	_updateTimer->start(intInterval, FALSE);
}

void Zega::stopGps() {
	_updateTimer->stop();
	_gpsLogHandler->stop();
#ifdef _DEBUG_PRINT
	cout << "Cardctl end." << endl;
#endif
	// zaurusڥᤷ
	enableSuspend();
	paintMaps();
}

void Zega::slotLoad() {
	QTKFileDialog dialog(this, "Load KML File");
	dialog.setFilter(tr("All|*.kml *.mrd *.log"));

	config.setGroup("Data");
	if (!config.hasKey("dir"))
		config.writeEntry("dir", "/mnt/card");

	dialog.setDirectory(config.readEntry("dir"));
	dialog.setDirectorySelectMode(false);
	dialog.setHomeDirectoryCdUpEnable(true);

	if (dialog.exec() == QDialog::Accepted) {
		if (dialog.getFileName().right(4) == ".log") {
			_gpsLogHandler->loadObsolateLogFile(dialog.getFileName());
			paintMaps();
			return;
		}

		int rtn = _route.loadFile(dialog.getFileName());
		if (rtn != 0) {
			if (rtn > 3 && rtn < 0) {
				rtn = 3;
			}
			QStringList msgs;
			msgs << tr("Could not load file.") << tr("File not found.") << tr("Unexpected error occurred.");
			QMessageBox::warning(this, tr("Load error"),
					msgs[rtn] + "\n" + dialog.getFileName());
			return;
		}

		// ̺
		paintMaps();
	}
}

void Zega::closeEvent(QCloseEvent *event) {
	int ret;
	if (!_gpsLogHandler->isAlive()) {
		ret = QMessageBox::warning(this, tr("Zegaλޤ"),
				tr("ZegaλƤɤǤ"),
				tr(""), tr("ᡪ"), QString::null,
				1, 1);
			
		if (ret == 0) {
			event->accept();
		} else {
			event->ignore();
		}
	} else {
		ret = QMessageBox::warning(this, tr("GPS¬Ǥ"),
				tr("GPS¬ΤZegaλǤޤ"),
				tr("OK"), QString::null, QString::null,
				1, 1);
		event->ignore();		
	}
}

void Zega::resizeEvent(QResizeEvent *event) {
	if (_mapBuffer->width() != this->width() || _mapBuffer->height()
			!= this->height()) {
		_mapBuffer->resize(this->width(), this->height());
		_drawGpsInfo->resize(this->width(), this->height());
		_drawMapInfo->resize(_mapBuffer->width(), _mapBuffer->height());
		paintMaps();
	}
}

void Zega::mousePressEvent(QMouseEvent *event) {
#ifdef _DEBUG_PRINT
	cout << "mousePressEvent " << event->pos().x() << ", " << event->pos().y() << endl;
#endif

	Radian tempRad;

	tempRad.setData(_center.getRadNS().toLong() + ((_pixelNS * (event->pos().y() - _mapBuffer->height()/2))>>_pixelBase));
	_center.setRadNS(tempRad);
	tempRad.setData(_center.getRadEW().toLong() + ((_pixelEW * (event->pos().x() - _mapBuffer->width()/2))>>_pixelBase));
	_center.setRadEW(tempRad);

	paintMaps();
}

void Zega::keyPressEvent(QKeyEvent *event) {
	Radian tempRad;
	int intWait = _cursorMoveDen;

	if (event->state() & ShiftButton) {
		intWait = _shiftCursorMoveDen;
	}

	switch (event->key()) {
	case Key_Up:
		tempRad = _center.getRadNS();
		tempRad.setData(tempRad.toLong() - (((_pixelNS * _mapBuffer->height())
				/intWait) >> _pixelBase));
		_center.setRadNS(tempRad);
		paintMaps();
		break;
	case Key_Down:
		tempRad = _center.getRadNS();
		tempRad.setData(tempRad.toLong() + (((_pixelNS * _mapBuffer->height())
				/intWait) >> _pixelBase));
		_center.setRadNS(tempRad);
		paintMaps();
		break;
	case Key_Left:
		tempRad = _center.getRadEW();
		tempRad.setData(tempRad.toLong() - (((_pixelEW * _mapBuffer->width())
				/intWait) >> _pixelBase));
		_center.setRadEW(tempRad);
		paintMaps();
		break;
	case Key_Right:
		tempRad = _center.getRadEW();
		tempRad.setData(tempRad.toLong() + (((_pixelEW * _mapBuffer->width())
				/intWait) >> _pixelBase));
		_center.setRadEW(tempRad);
		paintMaps();
		break;
	case Key_Space:
		if (_gpsLogHandler->isAlive()) {
			_center = _gpsLogHandler->getNowCoordinate();
			paintMaps();
		} else {
			return;
		}
		break;
	case Key_1:
		if (mapDirs.count() > 0)
			slotChScale(0);
		break;
	case Key_2:
		if (mapDirs.count() > 1)
			slotChScale(1);
		break;
	case Key_3:
		if (mapDirs.count() > 2)
			slotChScale(2);
		break;
	case Key_4:
		if (mapDirs.count() > 3)
			slotChScale(3);
		break;
	case Key_5:
		if (mapDirs.count() > 4)
			slotChScale(4);
		break;
	case Key_6:
		if (mapDirs.count() > 5)
			slotChScale(5);
		break;
	case Key_7:
		if (mapDirs.count() > 6)
			slotChScale(6);
		break;
	case Key_8:
		if (mapDirs.count() > 7)
			slotChScale(7);
		break;
	case Key_9:
		if (mapDirs.count() > 8)
			slotChScale(8);
		break;
	case Key_0:
		if (mapDirs.count() > 9)
			slotChScale(9);
		break;
	case Key_G:
		if (_gpsLogHandler->isStopped())
			startGps();
		else
			stopGps();
		break;
	case Key_C:
		slotToggleAutoCentered();
		break;
	case Key_F:
		slotFullScreen();
		break;
	case Key_H:
		slotToggleShowGpsLog();
		break;
	case Key_R:
		slotToggleShowRoute();
		break;
	case Key_L:
		slotLoad();
		break;
	case Key_S:
		slotSaveRoute();
		break;
	case Key_D:
		if (event->state() & ShiftButton) {
			_gpsLogHandler->changeDebugPrint();
		}
		slotToggleShowDestinationLine();
		break;
	case Key_A:
		slotAddRoutePoint();
		break;
	case Key_BackSpace:
	case Key_Delete:
		slotDelRoutePoint();
		break;
	case Key_O:
		slotReverseRoute();
		break;
	case Key_Z:
		slotCalcRemainLength();
		break;
	case Key_PageUp:
		tempRad = _center.getRadNS();
		tempRad.setData(tempRad.toLong() - ((_pixelNS *2) >> _pixelBase));
		_center.setRadNS(tempRad);
		paintMaps();
		break;
	case Key_PageDown:
		tempRad = _center.getRadNS();
		tempRad.setData(tempRad.toLong() + ((_pixelNS *2) >> _pixelBase));
		_center.setRadNS(tempRad);
		paintMaps();
		break;
	case Key_Home:
		tempRad = _center.getRadEW();
		tempRad.setData(tempRad.toLong() - ((_pixelEW *2) >> _pixelBase));
		_center.setRadEW(tempRad);
		paintMaps();
		break;
	case Key_End:
		tempRad = _center.getRadEW();
		tempRad.setData(tempRad.toLong() + ((_pixelEW *2) >> _pixelBase));
		_center.setRadEW(tempRad);
		paintMaps();
		break;
	case Key_Period:
		if (event->state() & ShiftButton) {
			slotHelp();
		}
		break;
	case Key_Escape:
		close();
		break;
	default:
		break;
	}
}

void Zega::slotCalcRemainLength() {
  _route.getRemain();
}

void Zega::slotToggleShowGpsLog() {
	if (_showGpsLog)
		_showGpsLog = false;
	else
		_showGpsLog = true;
	paintMaps();
}

void Zega::slotToggleShowRoute() {
	if (_showRoute)
		_showRoute = false;
	else
		_showRoute = true;
	paintMaps();
}

void Zega::slotToggleShowDestinationLine() {
	if (_showDestinationLine)
		_showDestinationLine = false;
	else
		_showDestinationLine = true;
	paintMaps();
}

void Zega::slotToggleAutoCentered() {
	if (_autoCentered)
		_autoCentered = false;
	else
		_autoCentered = true;
}

void Zega::slotAddRoutePoint() {
  cout << "add route" << endl;
	_route.addPoint(_center);
	cout << "add route point end" << endl;
	cout << _route._route.size() << endl;
	paintMaps();
  cout << "add route end" << endl;
}

void Zega::slotDelRoutePoint() {
	_route.deleteLastPoint();
	paintMaps();
}

void Zega::slotReverseRoute() {
	_route.reverse();
	paintMaps();
}

void Zega::slotSaveRoute() {
	QTKFileDialog dialog(this, "Save mrd File");
	dialog.setFilter(tr("All|*.mrd"));

	config.setGroup("Data");
	if (!config.hasKey("dir"))
		config.writeEntry("dir", "/mnt/card");

	dialog.setDirectory(config.readEntry("dir"));
	dialog.setDirectorySelectMode(false);
	dialog.setHomeDirectoryCdUpEnable(true);
	dialog.setMkDirButtonDisabled(false);
	dialog.setSaveMode(true);
	dialog.setSaveMode_ReWriteMessageDisabled(false);
	dialog.setHomeDirectoryCdUpEnable(true);

	if (dialog.exec() == QDialog::Accepted) {
		QString routeFileName = dialog.getFileName();
		if (dialog.getFileName().right(4) != ".mrd")
			routeFileName += ".mrd";

		QFile routeFile(routeFileName);
		if (routeFile.open(IO_WriteOnly)) {
			QTextStream s(&routeFile);
			for (vector<Coordinate>::iterator it = _route._route.begin(); it
					!= _route._route.end(); ++it) {
				s << it->getRadEW().toString2() << "," << it->getRadNS().toString2(true) << ",0\n";
			}

			routeFile.close();
		}
	}
}

void Zega::slotClearRoute() {
	int rtn = QMessageBox::warning(this, tr("Clear Route"),
			tr("롼Ⱦ̵äȤˤ㤦ɤ"), QMessageBox::Yes
					| QMessageBox::Default, QMessageBox::No
					| QMessageBox::Escape);
	if (rtn == QMessageBox::Yes) {
		_route.clear();
		paintMaps();
	}
}

void Zega::slotFullScreen() {
	// ȥСɽɽ
	showNormal();
	showFullScreen();
}

void Zega::paintEvent(QPaintEvent*) {
	if (_mapBuffer->width() == 0) {
		_mapBuffer->resize(this->width(), this->height());
		paintMaps();
	} else {
		bitBlt(this, 0, 0, _mapBuffer);
	}
}

// Ȥɽ
void Zega::slotHelp() {
	if(_help->isVisible())
		_help->hide();
	else
		_help->show();
}

// ץɽ
void Zega::about() {
	QMessageBox::about(
			this,
			tr("About this program"),
			tr("Zega Version 0.13.1\nThis is a GPS assistance program for Zaurus.\nCopyright HB1 2009.03.25"));
}

void Zega::paintMaps() {
	Radian radMapNS, radMapEW;
	Radian radMapWidth, radMapHeight;
	Radian radMinNS, radMinEW;
	QString mapFileName;
	QString tempStr;

	long cNS = _center.getRadNS().toLong();
	long cEW = _center.getRadEW().toLong();

	// ;
	_bsl = _mapBuffer->width();
	_bsu = _mapBuffer->height();
	_bsr = 0;
	_bsd = 0;

	// Ͽ޽񤭹ߥХåեꥢ
	_mapBuffer->fill();

	mapFileName = _mapList->searchMapFileName(cNS, cEW,
			&radMapEW, &radMapNS, &radMapWidth, &radMapHeight);

#ifdef _DEBUG_PRINT
	cout << "mapFileName: " << mapFileName << endl;
	cout << "mapNS: " << radMapNS.toNS() << endl;
	cout << "mapEW: " << radMapEW.toEW() << endl;
	cout << "mapMapHeight: " << radMapHeight.toString() << endl;
	cout << "mapMapWidth: " << radMapWidth.toString() << endl;
#endif

	if (!mapFileName.isEmpty()) {
		radMinEW.setData(radMapEW.toLong() - radMapWidth.toLong()/2);
		radMinNS.setData(radMapNS.toLong() - radMapHeight.toLong()/2);

		_buffer = _mapCache->getMap(mapFileName);
		if (_buffer != NULL) {
			int x = (cEW - radMinEW.toLong())* 1.0
					*(_buffer->width()) /radMapWidth.toLong();
			int y = (cNS - radMinNS.toLong())* 1.0
					*(_buffer->height()) /radMapHeight.toLong();
			int isx = x - _mapBuffer->width()/2;
			_bsl = 0;
			if (isx < 0) {
				_bsl = _mapBuffer->width()/2 - x;
				isx = 0;
			}
			int isy = y - _mapBuffer->height()/2;
			_bsu = 0;
			if (isy < 0) {
				_bsu = _mapBuffer->height()/2 - y;
				isy = 0;
			}

			// Ͽ޲ԡ
			bitBlt(_mapBuffer, _bsl, _bsu, _buffer, isx, isy, _buffer->width()
					- isx, _buffer->height() - isy);

			// ;׻
			_bsr = _bsl + _buffer->width() - isx;
			_bsd = _bsu + _buffer->height() - isy;
		}

#ifdef _DEBUG_PRINT
		cout << "compute pixel" << endl;
#endif
		// 1 pixel ٰ٤Ͽ
		int bh, bw;
		if (_buffer != NULL) {
			bh = _buffer->height();
			bw = _buffer->width();
		} else {
			bh = 500;
			bw = 500;
		}
		if (radMapWidth.toLong() < (+2147483647L >> PIXEL_BASE)
				&& radMapHeight.toLong() < (+2147483647L >> PIXEL_BASE)) {
			_pixelBase = PIXEL_BASE;
		} else {
			_pixelBase = 0;
		}

		_pixelNS = (radMapHeight.toLong() << _pixelBase) / bh;
		_pixelEW = (radMapWidth.toLong() << _pixelBase) / bw;
#ifdef _DEBUG_PRINT
		cout << "pixelNS: " << _pixelNS << endl;
		cout << "pixelEW: " << _pixelEW << endl;
#endif
	}

#ifdef _DEBUG_PRINT
	cout << "copy other maps" << endl;
#endif
	copyMap( 1, 0);
	copyMap(-1, 0);
	copyMap( 0, 1);
	copyMap( 0, -1);
	copyMap( 1, 1);
	copyMap( 1, -1);
	copyMap(-1, 1);
	copyMap(-1, -1);

	drawRoute();
	drawGpsLog();
	drawDestinationLine();

	// 롼ȻޤǤľΥɽ
	if(_route._route.size() > 0)
	  _drawMapInfo->repaint();

#ifdef _DEBUG_PRINT
	cout << "draw center cross" << endl;
#endif
	QPainter painter(_mapBuffer);
	painter.setPen(QPen(black, 1, SolidLine));
	painter.drawLine(_mapBuffer->width()/2 - 10, _mapBuffer->height()/2,
			_mapBuffer->width()/2 + 10, _mapBuffer->height()/2);
	painter.drawLine(_mapBuffer->width()/2, _mapBuffer->height()/2 - 10,
			_mapBuffer->width()/2, _mapBuffer->height()/2 + 10);
	bitBlt(this, 0, 0, _mapBuffer);
}

// 롼
void Zega::drawRoute() {
	if (!_showRoute)
		return;

	if (_route._route.size() <= 1)
	  return;

#ifdef _DEBUG_PRINT
	cout << "draw route" << endl;
#endif
	// ϰϤ١٤Ѵ
	long cNS = _center.getRadNS().toLong();
	long cEW = _center.getRadEW().toLong();
	int x, y, x2, y2;
	int cx = _mapBuffer->width()/2;
	int cy = _mapBuffer->height()/2;
	long lEW = cEW - ((_pixelEW * cx) >> _pixelBase);
	long uEW = cEW + ((_pixelEW * cx) >> _pixelBase);
	long lNS = cNS - ((_pixelNS * cy) >> _pixelBase);
	long uNS = cNS + ((_pixelNS * cy) >> _pixelBase);
	QPainter painter(_mapBuffer);

	// 롼Ȥ
	long lastEW, lastNS, nowEW, nowNS;
	long aEW, aNS, bEW, bNS;
	bool writeRoute = false;
	bool drawNewLine;
	painter.setPen(QPen(_routeColor, _routeWidth, Qt::SolidLine));
	vector<Coordinate>::iterator it = _route._route.begin();
	if (it != _route._route.end()) {
		lastEW = it->getRadEW().toLong();
		lastNS = it->getRadNS().toLong();
		it++;
		if (lastEW >= lEW && lastEW <= uEW && lastNS >= lNS && lastNS <= uNS) {
			writeRoute = true;
		}
	}
	while (it != _route._route.end()) {
		drawNewLine = false;
		nowEW = it->getRadEW().toLong();
		nowNS = it->getRadNS().toLong();
		if (nowEW >= lEW && nowEW <= uEW && nowNS >= lNS && nowNS <= uNS) {
			drawNewLine = true;
			writeRoute = true;
		} else if (writeRoute) {
			drawNewLine = true;
			writeRoute = false;
		} else {
			writeRoute = false;

			if (lastNS > nowNS) {
				aNS = nowNS;
				aEW = nowEW;
				bNS = lastNS;
				bEW = lastEW;
			} else {
				aNS = lastNS;
				aEW = lastEW;
				bNS = nowNS;
				bEW = nowEW;
			}
			if (aNS < lNS && bNS > lNS) {
				if (aEW < lEW) {
					if (bEW > lEW)
					drawNewLine = true;
				} else if (aEW < uEW) {
					drawNewLine = true;
				} else {
					if (bEW < uEW)
					drawNewLine = true;
				}
			}
			if (aNS < uNS && bNS > uNS) {
				if (bEW < lEW) {
					if (aEW > lEW)
					drawNewLine = true;
				} else if (bEW < uEW) {
					drawNewLine = true;
				} else {
					if (aEW < uEW)
					drawNewLine = true;
				}
			}

			if (drawNewLine == false) {
				if (lastEW > nowEW) {
					aNS = nowNS;
					aEW = nowEW;
					bNS = lastNS;
					bEW = lastEW;
				} else {
					aNS = lastNS;
					aEW = lastEW;
					bNS = nowNS;
					bEW = nowEW;
				}
				if (aEW < lEW && bEW > lEW) {
					if (aNS < lNS) {
						if (bNS > lNS)
						drawNewLine = true;
					} else if (aNS < uNS) {
						drawNewLine = true;
					} else {
						if (bNS < uNS)
						drawNewLine = true;
					}
				}
				if (aEW < uEW && bEW > uEW) {
					if (bNS < lNS) {
						if (aNS > lNS)
						drawNewLine = true;
					} else if (bNS < uNS) {
						drawNewLine = true;
					} else {
						if (aNS < uNS)
						drawNewLine = true;
					}
				}
			}
		}
		if (drawNewLine) {
			if (abs(lastEW - cEW) > abs(lastNS - cNS)) {
				if (lastEW < lEW) {
					aNS = lastNS - (1.0 * (lastNS - nowNS)) * (lastEW - lEW) / (lastEW - nowEW);
					x = ((lEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((aNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else if (lastEW > uEW) {
					aNS = lastNS - (1.0 * (lastNS - nowNS)) * (lastEW - uEW) / (lastEW - nowEW);
					x = ((uEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((aNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else {
					x = ((lastEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((lastNS - cNS) << _pixelBase) / _pixelNS + cy;
				}
			} else {
				if (lastNS < lNS) {
					aEW = lastEW - (1.0 * (lastEW - nowEW)) * (lastNS - lNS) / (lastNS - nowNS);
					x = ((aEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((lNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else if (lastNS > uNS) {
					aEW = lastEW - (1.0 * (lastEW - nowEW)) * (lastNS - uNS) / (lastNS - nowNS);
					x = ((aEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((uNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else {
					x = ((lastEW - cEW) << _pixelBase) / _pixelEW + cx;
					y = ((lastNS - cNS) << _pixelBase) / _pixelNS + cy;
				}
			}

			if (abs(nowEW - cEW) > abs(nowNS - cNS)) {
				if (nowEW < lEW) {
					aNS = nowNS - (1.0 * (nowNS - lastNS)) * (nowEW - lEW) / (nowEW - lastEW);
					x2 = ((lEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((aNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else if (nowEW > uEW) {
					aNS = nowNS - (1.0 * (nowNS - lastNS)) * (nowEW - uEW) / (nowEW - lastEW);
					x2 = ((uEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((aNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else {
					x2 = ((nowEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((nowNS - cNS) << _pixelBase) / _pixelNS + cy;
				}
			} else {
				if (nowNS < lNS) {
					aEW = nowEW - (1.0 * (nowEW - lastEW)) * (nowNS - lNS) / (nowNS - lastNS);
					x2 = ((aEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((lNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else if (nowNS > uNS) {
					aEW = nowEW - (1.0 * (nowEW - lastEW)) * (nowNS - uNS) / (nowNS - lastNS);
					x2 = ((aEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((uNS - cNS) << _pixelBase) / _pixelNS + cy;
				} else {
					x2 = ((nowEW - cEW) << _pixelBase) / _pixelEW + cx;
					y2 = ((nowNS - cNS) << _pixelBase) / _pixelNS + cy;
				}
			}
			painter.drawLine(x, y, x2, y2);
		}
		lastEW = nowEW;
		lastNS = nowNS;
		it++;
	}
}

// GPS
void Zega::drawGpsLog() {
	if (!_showGpsLog)
		return;

#ifdef _DEBUG_PRINT
	cout << "point obsolute logs" << endl;
#endif
	// ϰϤ١٤Ѵ
	Coordinate coordinate;
	long cNS = _center.getRadNS().toLong();
	long cEW = _center.getRadEW().toLong();
	long x, y;
	int cx = _mapBuffer->width()/2;
	int cy = _mapBuffer->height()/2;
	long lEW = cEW - ((_pixelEW * cx) >> _pixelBase);
	long uEW = cEW + ((_pixelEW * cx) >> _pixelBase);
	long lNS = cNS - ((_pixelNS * cy) >> _pixelBase);
	long uNS = cNS + ((_pixelNS * cy) >> _pixelBase);
	QPainter painter(_mapBuffer);

	// ľʳΥݥ
	painter.setPen(QPen(blue, 1, SolidLine));
	painter.setBrush(QBrush(_obsoluteLogColor, SolidPattern));
	_gpsLogHandler->setObsolateLogFirst();
	while (!_gpsLogHandler->endObsolateLog()) {
		coordinate = _gpsLogHandler->getNextObsolateLog();

		// ֤Ͽޤ˥ݥ
		if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
			x = ((coordinate.getRadEW().toLong() - cEW) << _pixelBase) / _pixelEW + cx;
			y = ((coordinate.getRadNS().toLong() - cNS) << _pixelBase) / _pixelNS + cy;
			painter.drawEllipse(x - _logWidth/2, y - _logWidth/2, _logWidth,
					_logWidth);
		}
	}

#ifdef _DEBUG_PRINT
	cout << "point recent logs" << endl;
#endif
	// ľΥݥ
	painter.setBrush(QBrush(_recentLogColor, SolidPattern));
	_gpsLogHandler->setRecentLogFirst();
	while (!_gpsLogHandler->endRecentLog()) {
		coordinate = _gpsLogHandler->getNextRecentLog();

		// ֤Ͽޤ˥ݥ
		if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
			x = ((coordinate.getRadEW().toLong() - cEW) << _pixelBase) / _pixelEW + cx;
			y = ((coordinate.getRadNS().toLong() - cNS) << _pixelBase) / _pixelNS + cy;
			painter.drawEllipse(x - _logWidth/2, y - _logWidth/2, _logWidth,
					_logWidth);
		}
	}
}

void Zega::drawDestinationLine() {
	if ((!_showDestinationLine) || _route._route.empty()) {
		return;
	}

#ifdef _DEBUG_PRINT
	cout << "draw destination line" << endl;
#endif
	long dEW = _route._route.back().getRadEW().toLong();
	long dNS = _route._route.back().getRadNS().toLong();
	long cNS = _center.getRadNS().toLong();
	long cEW = _center.getRadEW().toLong();

	int cx = _mapBuffer->width()/2;
	int cy = _mapBuffer->height()/2;
	long lEW = cEW - ((_pixelEW * cx) >> _pixelBase);
	long uEW = cEW + ((_pixelEW * cx) >> _pixelBase);
	long lNS = cNS - ((_pixelNS * cy) >> _pixelBase);
	long uNS = cNS + ((_pixelNS * cy) >> _pixelBase);

	QPainter painter(_mapBuffer);
	painter.setPen(QPen(_destinationLineColor, _destinationLineWidth, SolidLine));

	if (dEW >= lEW && dEW <= uEW && dNS >= lNS && dNS <= uNS) {
		int x = ((dEW - cEW) << _pixelBase) / _pixelEW + cx;
		int y = ((dNS - cNS) << _pixelBase) / _pixelNS + cy;
		painter.drawLine(cx, cy, x, y);
	} else {
		int x, y;
		getEdgePoint(&x, &y, dEW, dNS);
		painter.drawLine(cx, cy, x, y);
	}
}

void Zega::getEdgePoint(int *x, int *y, long dEW, long dNS) {
	long cNS = _center.getRadNS().toLong();
	long cEW = _center.getRadEW().toLong();

	if (abs(dEW - cEW) < abs(dNS - cNS)) {
		double rate = (dEW - cEW) * 1.0 * _pixelNS / (abs(dNS - cNS) * 1.0
				* _pixelEW);
		if (dNS - cNS > 0)
			*y = _mapBuffer->height();
		else
			*y = 0;
		*x = int(rate * _mapBuffer->height() / 2 + _mapBuffer->width() / 2);
	} else {
		double rate = (dNS - cNS) * 1.0 * _pixelEW / (abs(dEW - cEW) * 1.0
				* _pixelNS);
		if (dEW - cEW > 0)
			*x = _mapBuffer->width();
		else
			*x = 0;
		*y = int(rate * _mapBuffer->width() / 2 + _mapBuffer->height() / 2);
	}
}

// Ͽɹ
bool Zega::copyMap(int x, int y) {
	Radian radMapEW, radMapNS;
	Radian radMapWidth, radMapHeight;
	QString mapFileName;

	if (x < 0 && _bsl <= 0)
		return false;
	if (x > 0 && _bsr >= _mapBuffer->width())
		return false;
	if (y < 0 && _bsu <= 0)
		return false;
	if (y > 0 && _bsd >= _mapBuffer->height())
		return false;

	mapFileName = _mapList->pointMapFileName(y, x, &radMapEW, &radMapNS,
			&radMapWidth, &radMapHeight);
	if (!mapFileName.isEmpty()) {
		_buffer = _mapCache->getMap(mapFileName);
		if (_buffer == NULL) {
			return false;
		}

		long cNS = _center.getRadNS().toLong();
		long cEW = _center.getRadEW().toLong();
		long minNS = radMapNS.toLong() - radMapHeight.toLong()/2;
		long minEW = radMapEW.toLong() - radMapWidth.toLong()/2;
		int w = (cEW - minEW)*(_buffer->width())/radMapWidth.toLong();
		int h = (cNS - minNS)*(_buffer->height())/radMapHeight.toLong();

		int isx = w - _mapBuffer->width()/2;
		int bsx = 0;
		if (isx < 0) {
			bsx = _mapBuffer->width()/2 - w;
			isx = 0;
		}
		int isy = h - _mapBuffer->height()/2;
		int bsy = 0;
		if (isy < 0) {
			bsy = _mapBuffer->height()/2 - h;
			isy = 0;
		}

		int copy_width = _buffer->width() - isx;
		if (x > 0 && _bsr > bsx) {
			isx += _bsr - bsx;
			copy_width -= _bsr - bsx;
			bsx = _bsr;
		} else if (x < 0 && copy_width > _bsl - bsx)
{
			copy_width = _bsl - bsx;
		}

		int copy_height = _buffer->height() - isy;
		if (y > 0 && _bsd > bsy) {
			isy += _bsd - bsy;
			copy_height -= _bsd - bsy;
			bsy = _bsd;
		} else if (y < 0 && copy_height > _bsu - bsy)
{
			copy_height = _bsu - bsy;
		}

		// Ͽ޲ԡ
		bitBlt(_mapBuffer, bsx, bsy, _buffer, isx, isy, copy_width,
					copy_height);
		//  _bsl, _bsr, _bsu, _bsd ׻
	}
	return true;
}
